import os
import sys
import traceback
from os.path import join, isfile

import aspen
from aspen import gauntlet, resources, sockets
from aspen.http.request import Request
from aspen.http.response import Response
from aspen.configuration import Configurable


class Website(Configurable):
    """Represent a website.

    This object holds configuration information, and also knows how to start
    and stop a server, *and* how to handle HTTP requests (per WSGI). It is
    available to user-developers inside of their resources and hooks.

    """

    def __init__(self, argv=None):
        """Takes an argv list, without the initial executable name.
        """
        self.configure(argv)
    
    def __call__(self, environ, start_response):
        """WSGI interface.
        """
        request = Request.from_wsgi(environ) # too big to fail :-/
        response = self.handle(request)
        response.request = request # Stick this on here at the last minute
                                   # in order to support close hooks.
        return response(environ, start_response)

    def start(self):
        aspen.log("Starting up Aspen website.")
        self.hooks.startup.run(self)
        self.network_engine.start()

    def stop(self):
        aspen.log("Shutting down Aspen website.")
        self.hooks.shutdown.run(self)
        self.network_engine.stop()

    def handle(self, request):
        """Given an Aspen request, return an Aspen response.

        Aspen uses Resource subclasses to generate responses. See
        aspen/resources/__init__.py.

        """
        try:
            try:
                #self.copy_configuration_to(request)
                request.website = self

                self.hooks.inbound_early.run(request)
                gauntlet.run(request) # sets request.fs
                request.socket = sockets.get(request)
                self.hooks.inbound_late.run(request)

                # Look for a Socket.IO socket (http://socket.io/).
                if isinstance(request.socket, Response):    # handshake
                    response = request.socket
                    request.socket = None
                elif request.socket is None:                # non-socket
                    request.resource = resources.get(request)
                    response = request.resource.respond(request)
                else:                                       # socket
                    response = request.socket.respond(request)
            except:
                response = self.handle_error_nicely(request)
        except Response, response:
            # Grab the response object in the case where it was raised.  In the
            # case where it was returned, response is set in a try block above.
            pass
        else:
            # If the response object is coming from handle_error via except 
            # Response, then it already has request on it and the early hooks
            # have already been run. If it fell off the edge un-exceptionally,
            # we need to take care of those two things.
            response.request = request
            self.hooks.outbound_early.run(response)

        self.hooks.outbound_late.run(response)
        self.log_access(request, response) # TODO is this at the right level?
        return response

    def handle_error_nicely(self, request):
        """Try to provide some nice error handling.
        """
        try:                        # nice error messages
            tb_1 = traceback.format_exc()
            response = sys.exc_info()[1]
            if not isinstance(response, Response):
                aspen.log_dammit(tb_1)
                response = Response(500, tb_1)
            elif 200 <= response.code < 300:
                return response
            response.request = request
            self.hooks.outbound_early.run(response)
            fs = self.ours_or_theirs(str(response.code) + '.html')
            if fs is None:
                fs = self.ours_or_theirs('error.html')
            if fs is None:
                raise
            request.fs = fs
            request.original_resource = request.resource
            request.resource = resources.get(request)
            response = request.resource.respond(request, response)
            return response
        except Response, response:  # no nice error template available
            raise
        except:                     # last chance for tracebacks in the browser
            tb_2 = traceback.format_exc().strip()
            tbs = '\n\n'.join([tb_2, "... while handling ...", tb_1])
            aspen.log_dammit(tbs)
            if self.show_tracebacks:
                response = Response(500, tbs)
            else:
                response = Response(500)
            response.request = request
            raise response

    def find_ours(self, filename):
        """Given a filename, return a filepath.
        """
        return join(os.path.dirname(__file__), 'www', filename)

    def ours_or_theirs(self, filename):
        """Given a filename, return a filepath or None.
        """
        if self.project_root is not None:
            theirs = join(self.project_root, filename)
            if isfile(theirs):
                return theirs

        ours = self.find_ours(filename)
        if isfile(ours):
            return ours

        return None

    def log_access(self, request, response):
        """Log access. With our own format (not Apache).
        """

        if self.quiet_level > 0:
            return


        # What was the URL path translated to?
        # ====================================

        fs = request.fs
        if fs.startswith(self.www_root):
            fs = fs[len(self.www_root):]
            if fs:
                fs = '.'+fs
        else:
            fs = '...' + fs[-21:]
        msg = "%-24s %s" % (request.line.uri.path.raw, fs)


        # Where was response raised from?
        # ===============================

        tb = sys.exc_info()[2]
        if tb is not None:
            while tb.tb_next is not None:
                tb = tb.tb_next
            frame = tb.tb_frame
            filename = tb.tb_frame.f_code.co_filename.split(os.sep)[-1]
            response = "%s (%s:%d)" % (response, filename, frame.f_lineno)
        else:
            response = str(response)


        # Log it.
        # =======

        aspen.log("%-36s %s" % (response, msg))



"""Aspen supports Socket.IO sockets. http://socket.io/
"""
from aspen.resources.dynamic_resource import DynamicResource


class SocketResource(DynamicResource):

    min_pages = 1
    max_pages = 4

    def respond(self):
        """Override and kill it. For sockets the Socket object responds.
        """
        raise NotImplemented

    def compile_third(self, one, two, three, padding):
        """Given four bytestrings, return a code object.

        This method depends on self.fs.

        """
        # See DyanmicResource._compile for comments on this algorithm.
        three = three.replace('\r\n', '\n')
        three = padding + three
        three = compile(three, self.fs, 'exec')
        return three

    def compile_fourth(self, one, two, three, four, padding):
        """Given five bytestrings, return a code object.

        This method depends on self.fs.

        """
        # See DyanmicResource._compile for comments on this algorithm.
        four = four.replace('\r\n', '\n')
        four = padding + four
        four = compile(four, self.fs, 'exec')
        return four

    def exec_second(self, socket, request):
        """Given a Request, return a context dictionary.
        """
        context = request.context
        context.update(self.one)
        context['socket'] = socket
        context['channel'] = socket.channel
        exec self.two in context
        return context 



import optparse

import aspen


# OptionParser
# ------------

usage = "aspen [options]"


version = """\
aspen, version %s

(c) 2006-2012 Chad Whitacre and contributors
http://aspen.io/
""" % aspen.__version__


description = """\
Aspen is a Python web framework. By default this program will start serving a
website from the current directory on port 8080. Options are as follows. See 
also http://aspen.io/.
"""


class DEFAULT(object):
    def __repr__(self):
        return "<DEFAULT>"
DEFAULT = DEFAULT()


def OptionParser():
    optparser = optparse.OptionParser( usage=usage
                                     , version=version
                                     , description=description
                                      )

    basic = optparse.OptionGroup(optparser, "Basic Options")

    basic.add_option( "-f", "--configuration_scripts"
                    , help=("comma-separated list of paths to configuration "
                            "files in Python syntax to exec in addition to "
                            "$ASPEN_PROJECT_ROOT/configure-aspen.py")
                    , default=DEFAULT
                     )
    basic.add_option( "-a", "--network_address"
                    , help=("the IPv4, IPv6, or Unix address to bind to "
                            "[0.0.0.0:8080]")
                    , default=DEFAULT
                     )
    basic.add_option( "-e", "--network_engine"
                    , help=( "the HTTP engine to use, one of "
                           + "{%s}" % ','.join(aspen.ENGINES)
                           + " [%s]" % aspen.ENGINES[1]
                            )
                    , default=DEFAULT
                     )
    basic.add_option( "-q", "--quiet_level"
                    , help=("a small integer; 1 will suppress most of aspen's "
                            "internal logging, 2 will suppress all it [0]")
                    , default=DEFAULT
                     )
    basic.add_option( "-p", "--project_root"
                    , help=("the filesystem path of the directory in "
                            "which to look for project files like "
                            "template bases and such, relative to "
                            "$ASPEN_WWW_ROOT []")
                    , default=DEFAULT
                     )
    basic.add_option( "-w", "--www_root"
                    , help=("the filesystem path of the document "
                            "publishing root [.]")
                    , default=DEFAULT
                     )


    extended = optparse.OptionGroup( optparser, "Extended Options"
                                   , "I judge these variables to be less-"
                                     "often configured from the command "
                                     "line. But who knows?"
                                    )
    extended.add_option( "--changes_kill"
                       , help=("if set to yes/true/1, changes to files in "
                               "$ASPEN_PROJECT_ROOT will cause aspen to exit "
                               "[no]")
                       , default=DEFAULT
                        )
    extended.add_option( "--charset_dynamic"
                       , help=("this is set as the charset for template "
                               "resources of Content-Type text/* [UTF-8]")
                       , default=DEFAULT
                        )
    extended.add_option( "--charset_static"
                       , help=("if set, this will be sent as the charset for "
                               "static resources of Content-Type text/*; if "
                               "you want to punt and let browsers guess, then "
                               "just leave this unset []")
                       , default=DEFAULT
                        )
    extended.add_option( "--indices"
                       , help=("a comma-separated list of filenames to look "
                               "for when a directory is requested directly; "
                               "prefix with + to extend previous "
                               "configuration instead of overriding "
                               "[index.html, index.json]")
                       , default=DEFAULT
                        )
    extended.add_option( "--list_directories"
                       , help=("if set to {yes,true,1}, aspen will serve a "
                               "directory listing when no index is available "
                               "[no]")
                       , default=DEFAULT
                        )
    extended.add_option( "--media_type_default"
                       , help=("this is set as the Content-Type for resources "
                               "of otherwise unknown media type [text/plain]")
                       , default=DEFAULT
                        )
    extended.add_option( "--media_type_json"
                       , help=("this is set as the Content-Type of JSON "
                               "resources [application/json]")
                       , default=DEFAULT
                        )
    extended.add_option( "--show_tracebacks"
                       , help=("if set to {yes,true,1}, 500s will have a "
                               "traceback in the browser [no]")
                       , default=DEFAULT
                        )


    optparser.add_option_group(basic)
    optparser.add_option_group(extended)
    return optparser



from aspen import Response
from aspen.resources.resource import Resource


PAGE_BREAK = chr(12)

# Global page limits. There are further limits per-resource-type.
MIN_PAGES = 2
MAX_PAGES = 4

class StringDefaultingList(list):
    def __getitem__(self, key):
        try:
            return super(StringDefaultingList, self).__getitem__(key)
        except KeyError:
            return str(key)
ORDINALS = StringDefaultingList([ 'zero' , 'one' , 'two', 'three', 'four'
                                , 'five', 'six', 'seven', 'eight', 'nine'])


class DynamicResource(Resource):
    """This is a base class for json, socket, and template resources.
    """

    min_pages = None # set on subclass
    max_pages = None
    
    def __init__(self, *a, **kw):
        assert MIN_PAGES <= self.max_pages <= MAX_PAGES # sanity check
        super(DynamicResource, self).__init__(*a, **kw)
        pages = self._parse(self.raw)
        self.one, self.two, self.three, self.four = self._compile(*pages)

    def respond(self, request, response=None):
        """Given a Request and maybe a Response, return or raise a Response.
        """
        response = response or Response()

       
        # Populate context.
        # =================
        
        context = request.context
        context.update(self.one)
        context['response'] = response
        context['resource'] = self
   

        # Exec the script.
        # ================
    
        try:
            exec self.two in context 
        except Response, response:
            response = self.process_raised_response(response)
            raise response


        # Hook.
        # =====

        return self.get_response(context)


    def _parse(self, raw):
        """Given a bytestring, return a list of four items.
        
        If there are too few pages, raise AssertionError. Any resource with
        only one page should land in StaticResource, not here.
        
        If there are too many pages, raise SyntaxError. 

        If there are fewer than self.max_pages, then pad the front of the list
        with up to two empty strings, and the end with up to one.

        If self.max_pages is less than four (i.e., three or two), pad the end
        of the list with None.
        
        """

        # Support caret-L in addition to .
        uncareted = raw.replace("^L", PAGE_BREAK)
        pages = uncareted.split(PAGE_BREAK)
        npages = len(pages)

        # Check for too few pages.
        assert npages >= self.min_pages # sanity check; bug if False

        # Check for too many pages.
        if npages > self.max_pages:     # user error if True
            type_name = self.__class__.__name__[:-len('resource')]
            msg = "%s resources must have exactly %s pages; %s has %s."
            msg %= ( type_name
                   , ORDINALS[self.max_pages]
                   , self.fs
                   , ORDINALS[npages]
                    )
            raise SyntaxError(msg)

        # Pad front (and back?) with empty strings.
        while len(pages) < (self.max_pages - 1):
            pages.insert(0, '')
        if len(pages) < self.max_pages:
            pages.append('')


        # Pad the back with None.
        while len(pages) < MAX_PAGES:
            pages.append(None)

        return pages
       
    def _compile(self, one, two, three, four):
        """Given four items, return a 4-tuple of compiled objects.

        All dynamic resources compile the first two pages the same way. It's
        the third page that differs, so we require subclasses to provide a hook
        for that.

        """

        # Standardize newlines.
        # =====================
        # compile requires \n, and doing it now makes the next line easier.

        one = one.replace('\r\n', '\n')
        two = two.replace('\r\n', '\n')
        if three is not None:
            three = three.replace('\r\n', '\n')


        # Compute paddings and pad the second and third pages.
        # ====================================================
        # This is so we get accurate tracebacks. We will pass padding_* to the
        # compile_* hooks in case subclasses want to use them.
        
        padding = lambda s: ''.join(['\n' for n in range(s.count('\n'))])
        padding_two = padding(one)
        padding_three = padding_two + padding(two)
        two = padding_two + two 


        # Exec the first page and compile the second.
        # ===========================================
        # Below in render we take care not to mutate context.

        context = dict()
        context['__file__'] = self.fs
        context['website'] = self.website
        
        one = compile(one, self.fs, 'exec')
        two = compile(two, self.fs, 'exec')

        exec one in context
        one = context


        # Third and Fourth
        # ================

        three = self.compile_third(one, two, three, padding_two)
        four = self.compile_fourth(one, two, three, four, padding_three)


        return one, two, three, four


    # Hooks
    # =====

    def compile_third(self, *a):
        """Given a bytestring, return an object.
        """
        raise NotImplementedError

    def compile_fourth(self, *a):
        """Given a bytestring, return an object.
        """
        raise NotImplementedError

    def process_raised_response(self, response):
        """Given a response object, return a response object.
        """
        return response

    def get_response(self, context):
        """Given a context dictionary, return a Response object.
        """
        raise NotImplementedError



import os
import sys
import threading
import time

try:                # Python >= 2.6
    from collections import Callable
    def is_callable(obj):
        return isinstance(obj, Callable)
except ImportError: # Python < 2.6
    from operator import isCallable as is_callable

# imports of convenience
from aspen.http.response import Response
from aspen import json_ as json
Response, json # Shut up, PyFlakes. I know I'm dependent on you.


__version__ = "~~VERSION~~"
WINDOWS = sys.platform[:3] == 'win'
ENGINES = ['cheroot', 'cherrypy', 'diesel', 'eventlet', 'gevent', 'pants', 
           'rocket', 'tornado', 'twisted']


def thrash():
    """This is a very simple tool to restart a process when it dies.

    It's designed to restart aspen in development when it dies because files
    have changed and you set changes_kill to 'yes' in the [aspen.cli] section
    of aspen.conf.

    http://aspen.io/thrash/

    """
    try:
        while 1:
            os.system(' '.join(sys.argv[1:]))
            time.sleep(1)
    except KeyboardInterrupt:
        pass


# Logging. It's simple.
# =====================
# The level attribute is set in configuration/__init__.py. If you pass -q level
# will be set to 1 and only really important things (tracebacks) will be
# logged. If you pass -qq it will be set to 2 and nothing will be logged.

LOG_LEVEL = 0

_pid = os.getpid()
def log(message, level=0):
    message = unicode(message).encode('UTF-8', 'backslashreplace') # XXX buggy?
    if level >= LOG_LEVEL:
        t = threading.current_thread()
        for line in message.splitlines():  # doesn't include linebreaks 
            print "pid-%d thread-%d (%s) %s" % (_pid, t.ident, t.name, line)
        sys.stdout.flush()

def log_dammit(message):
    log(message, 1)



class ConfigurationError(StandardError):
    """This is an error in any part of our configuration.
    """

    def __init__(self, msg):
        StandardError.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg



import codecs


# Register a 'repr' error strategy.
# =================================
# Sometimes we want to echo bytestrings back to a user, and we don't know what 
# encoding will work. This error strategy replaces non-decodable bytes with 
# their Python representation, so that they are human-visible.
# 
# See also:
#   - https://github.com/dcrosta/mongo/commit/e1ac732
#   - http://www.crummy.com/software/BeautifulSoup/bs4/doc/#unicode-dammit

def replace_with_repr(unicode_error):
    offender = unicode_error.object[unicode_error.start:unicode_error.end]
    return (unicode(repr(offender).strip("'").strip('"')), unicode_error.end)

codecs.register_error('repr', replace_with_repr)


def unicode_dammit(s, encoding="UTF-8"):
    """Given a bytestring, return a unicode decoded with `encoding`.

    Any bytes not decodable with UTF-8 will be replaced with their Python
    representation, so you'll get something like u"foo\\xefbar".

    """
    if not isinstance(s, str):
        raise TypeError("You gave me something besides a str.")
    errors = 'repr'
    return s.decode(encoding, errors)

def ascii_dammit(s):
    """Given a bytestring, return a bytestring.

    The returned bytestring will have any non-ASCII bytes replaced with
    their Python representation, so it will be pure ASCII.

    """
    return unicode_dammit(s, encoding="ASCII").encode("ASCII")



import os

from aspen.engines import ThreadedBuffer
from aspen.http.request import Request
from aspen.sockets.channel import Channel
from aspen.sockets.socket import Socket
from aspen.website import Website
from aspen.testing import fix


def make_request(filename='echo.sock'):
    request = Request(uri='/echo.sock')
    request.website = Website([])
    request.fs = fix(filename)
    return request

def make_socket(filename='echo.sock', channel=None):
    request = make_request(filename='echo.sock')
    if channel is None:
        channel = Channel(request.line.uri.path.raw, ThreadedBuffer)
    socket = Socket(request, channel)
    return socket

class SocketInThread(object):

    def __enter__(self, filename='echo.sock'):
        self.socket = make_socket(filename) 
        self.socket.loop.start()
        return self.socket

    def __exit__(self, *a):
        self.socket.loop.stop()
